"""
Copyright 2011,  Leif Theden & Michael Bachmann


This file is part of Nein.

Nein is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Nein is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Nein.  If not, see <http://www.gnu.org/licenses/>.
"""

import pygame
from pygame.locals import *
from collections import deque

# sub pixel rendering from
# http://www.willmcgugan.com/blog/tech/2007/4/25/going-sub-pixel-with-pygame/
# not implemented...

# kinda the main processor of game intereaction
class StateDriver(object):
    def __init__(self, parent):
        self.parent = parent
        self._states = deque()
        self.reloadScreen()

    def reloadScreen(self):
        """ Called when the display changes mode somehow. """
        self._screen = self.parent.get_screen()

    def done(self):
        self.getCurrentState().deactivate()
        self._states.pop()
        state = self.getCurrentState()
        if state.activated:
            state.reactivate()
        else:
            state.activate()

    def getCurrentState(self):
        try:
            return self._states[-1]
        except:
            self.quit()

    def get_size(self):
        return self._screen.get_size()

    def quit(self):
        pygame.quit()

    def replace(self, state):
        self.getCurrentState().deactivate()
        self._states.pop()
        self.start(state)

    def start(self, state):
        self._states.append(state)
        self.getCurrentState().activate()

    def push(self, state):
        self._states.appendleft(state)

    def run(self):
        currentState = self.getCurrentState()
        import lib.gfx as gfx

        # deref for speed
        event_poll = pygame.event.poll
        event_pump = pygame.event.pump
        wait       = pygame.time.wait
        get_ticks  = pygame.time.get_ticks
        current_state = self.getCurrentState

        while currentState:
            event = event_poll()

            while event.type != NOEVENT:
                if event.type == QUIT:
                    currentState = None
                    break

                elif event.type == KEYUP or event.type == KEYDOWN:
                    if event.key == K_ESCAPE:
                        currentState = None
                        break
                    else:
                        currentState.handle_event(event) 
                else:
                    currentState.handle_event(event)

                event = event_poll()

            if currentState:
                currentState.update(get_ticks())
                currentState.draw(self._screen)
                currentState = current_state()
                gfx.update_display()

class UpdateableObject(object):
    def __init__(self):
        self._is_enabled = True
        self.init()

    # just a hook for subclasses
    def init(self):
        pass

    def update(self, time):
        pass

    def enable(self):
        self._is_enabled = True

    def disable(self):
        self._is_enabled = False

class DrawableObject(UpdateableObject):
    def __init__(self):
        super(DrawableObject, self).__init__()
        self._is_visible = True
        self.dirty = False

    def draw(self, surface):
        raise NotImplementedError, self

    def show(self):
        self.is_visible = True

    def hide(self):
        self._is_visible = False

class KeyableObject(object):
    def __init__(self, keys=None):
        self.keys = keys

    def mask_event(self, key, unicode, pressed):
        if self.keys:
            if key not in self.keys:
                return
        self.key_event(key, unicode, pressed)

    def keyEvent(self,key,unicode, pressed):
        raise SubclassShouldImplement

class EventHandler(object):
    """
    Events from PyGame with trickle down through listeners
    based on order of insertion.
    Make sure to return the event if you want to all lesser
    Listeners to use it as well.  Return None to kill it.
    """

    def handle_event(self, event):
        return event

class GameState(object):
    def __init__(self, driver):
        self._driver = driver
        self.activated = False

    # when focused for first time
    def activate(self):
        pass

    # when regaining focus after having lost it
    def reactivate(self):
        pass

    # when losing focus
    def deactivate(self):
        pass

    def draw(self, surface):
        pass

    def update(self, time):
        pass

class World(DrawableObject, EventHandler):
    """
    Handles ALL SORTS OF STUFF (even if it looks like it doesn't).
    """

    def __init__(self):
        DrawableObject.__init__(self)

        self._drawables = []
        self._updateables = []
        self._eventhandlers = []

    def _add_other(self, obj):
        raise ValueError, "This type of object is not handled: %s" % obj.__class__

    def add(self, obj):
        added = False

        if isinstance(obj, EventHandler):
            self._eventhandlers.append(obj)
            added = True

        if isinstance(obj, DrawableObject):
            self._drawables.append(obj)
            added = True

        if isinstance(obj, UpdateableObject):
            self._updateables.append(obj)
            added = True

        if isinstance(obj, list):
            [ self.add(i) for i in obj ]
            added = True

        if not added:
            self._add_other(obj)

    def handle_event(self, event):
        [ obj.handle_event(event) for obj in self._eventhandlers ]

    def all_of_type(self, t):
        return [ i for i in self._updateables if type(i) == i ]

    def remove(self, obj):
        # not optimized.
        if obj in self._updateables: self._updateables.remove(obj)
        if obj in self._drawables: self._drawables.remove(obj)
        if obj in self._eventhandlers: self._eventhandlers.remove(obj)

    def update(self, time):
        to_remove = []
        [ to_remove.append(obj) for obj in self._updateables if obj.update(time) == False ]
        [ self.remove(obj) for obj in to_remove ]

    def draw(self, surface):
        [ obj.draw(surface) for obj in self._drawables ]
