"""
Copyright 2011,  Leif Theden & Michael Bachmann


This file is part of Nein.

Nein is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Nein is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Nein.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
The idea of this class is to support dynamic and interesting animations
in a GUI and to use the physics engine to do the work.
"""

from pygame.rect import Rect
import pygame
from engine import DrawableObject


class GuiElement(DrawableObject):
    pass

class PhysicsGuiElement(object):
    """
    This special class has a physics body associated with it.
    The body should be somewhere it would not normally be seen.
    It has a special collision group so that it doesn't
    interact with objects in the simulation.
    It is used for dynamic animations in a GUI.
    """

    def update_rect(self, camera):
        """
        the rect should be fixed in one position, regardless of the camera
        """
        pass


class HorizBar(GuiElement):
    """
    This GuiElement is a horizontal bar.  Not that interesting.
    Well, it kinda is.

    IF you pass a func to it, then when it updates, it will call the func
    on every draw to get the new value.
    Also, if you set "ocd" higher than -1,
    it will check the func every x seconds (or whatever your time is measured in),
    ocd defualt is 0, so it will check on every update.
    which is up to you to make sure that happens.  decisions, decisions...

    so, if you decided to have it check, make sure the func passed is fast
    """

    def __init__(self, rect, value=0.0, minimum=0.0, maximum=1.0, func=None, ocd=0):
        GuiElement.__init__(self)
        rect = Rect(rect)
        self.value = value
        self.minimum = minimum
        self.maximum = maximum
        self.rect = rect
        self.func = func
        self.ocd = ocd
        self.image = pygame.surface.Surface(rect.size)
        self.dirty = True
        self.last_update = 0

        if self.ocd < 0: self.ocd = 0

        # colors, if you were wondering
        self.fg = (255,0,255)
        self.bg = (0,0,0)

    def update(self, time):
        if self.ocd == 0:
            self.value = self.func()
        
        elif self.last_update + self.ocd <= time:
            self.value = self.func()
            self.last_update = time

    def draw(self, surface):
        if self.dirty:
            x, y = self.rect.size
            split = self.value / self.maximum * x
            self.image.fill(self.fg, (0,0,split,y))
            self.image.fill(self.bg, (split,0,x,y))

        surface.blit(self.image, self.rect)
