"""
Copyright 2011,  Leif Theden & Michael Bachmann


This file is part of Nein.

Nein is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Nein is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Nein.  If not, see <http://www.gnu.org/licenses/>.
"""

import pygame
from pygame.rect import Rect
from pygame.draw import aalines, rect
from pygame.sprite import OrderedUpdates
from actor import Actor
from sprite import PhysicsSprite, SpriteBundle
from engine import EventHandler, UpdateableObject, DrawableObject, World
from math import cos, pi, sqrt
from random import uniform, randint
import Box2D as b2d
import sfx, gfx

from pygame.locals import *

dbg_outlines = False


BOX2DTIMESTEP = 30.0

"""
i have a very precarious system of tracking shapes and bodies.
since shapes are what the collions module returns, i need
to track the actors based on shapes, which is lame.

so a dict is modified when the sprite is added.
just found out you can shape.GetBody()...oh well
"""


"""
very simple collision detection system and sounds for it
a combination of collision groups will cause a sound to play
a better system would be tracking bodies, but this is much easier
"""
c_sound = {}


# sound idea: have a pulse wave generator running for bodies moving quickly

class fwContactTypes:
    """
    Acts as an enum, holding the types necessary for contacts:
    Added, persisted, and removed
    """
    contactUnknown = 0
    contactAdded = 1
    contactPersisted = 2
    contactRemoved = 3

class fwContactPoint:
    """
    Structure holding the necessary information for a contact point.
    All of the information is copied from the contact listener callbacks.
    """
    shape1 = None
    shape2 = None
    normal = None
    position = None
    velocity = None
    id  = None
    state = 0

class fwContactListener(b2d.b2ContactListener):
    """
    Handles all of the contact states passed in from Box2D.

    """
    test = None
    def __init__(self):
        super(fwContactListener, self).__init__()

    def handleCall(self, state, point):
        if not self.test: return

        cp          = fwContactPoint()
        cp.shape1   = point.shape1
        cp.shape2   = point.shape2
        cp.position = point.position.copy()
        cp.normal   = point.normal.copy()
        cp.velocity = point.velocity.copy()
        cp.id       = point.id
        cp.state    = state
        self.test.points.append(cp)

    def Add(self, point):
        self.handleCall(fwContactTypes.contactAdded, point)

    def Persist(self, point):
        self.handleCall(fwContactTypes.contactPersisted, point)

    def Remove(self, point):
        self.handleCall(fwContactTypes.contactRemoved, point)

class PhysicsWorld(World):
    """
    Game object that is a physics world.  Fancy that.
    Don't forget to add a camera.
    """

    def __init__(self, lower=(-100, -100), upper=(200,200), gravity=(0.0, -10.0), scale=10.0):
        World.__init__(self)

        worldAABB=b2d.b2AABB()
        worldAABB.lowerBound = lower
        worldAABB.upperBound = upper
        gravity = (0.0, -10.0)
        doSleep = True

        self.world = b2d.b2World(worldAABB, gravity, doSleep)

        self.timeStep = 1.0 / BOX2DTIMESTEP
        self.velocityIterations = 10
        self.positionIterations = 8

        self.contactListener = fwContactListener()
        self.contactListener.test = self
        self.world.SetContactListener(self.contactListener)

        self.points = []
        self.stale_sounds = []

        # scaling objects makes the engine faster
        self.ppm = scale

        # kinda redundant...oh well
        self._sprites = []
        self._shapeDB = {}

        # default sound for explosions
        self.snd_ex = sfx.load("CatSplosion.wav")

        # default sound for high velocity impacts
        s = ["bighit0.wav", "bighit1.wav", "bighit2.wav", "bighit3.wav", "bighit4.wav"]
        self.snd_bump_high = [ sfx.load(i) for i in s ]

        # default sounds for low velocity impacts
        s = ["smallhit0.wav", "smallhit1.wav", "smallhit2.wav"]
        self.snd_bump_low = [ sfx.load(i) for i in s ]

        # HACK
        global c_sound
        c_sound[0,-1] = True
        c_sound[-1,0] = True
        c_sound[0,2] = True
        c_sound[2,0] = True

    def explode(self, sb):
        # first we destroy all the joints
        [ self.world.DestroyJoint(j) for j in sb.joints.values() ]

        # next, calculate a force that will move the pieces
        mass = sum([s.body.massData.mass for s in sb.sprites])
        f = mass / len(sb.sprites) * 8

        # then we apply a impulse in a random direction on each body in the bundle
        [ s.body.ApplyImpulse(
            b2d.b2Vec2(uniform(-f,f), uniform(-f,f)), \
            s.body.GetWorldCenter()) for s in sb.sprites ]

        # tune down the bounciness of the parts to prevent a crazy situation
        for shape in [ k for k,v in self._shapeDB.items() if v == sb ]:
            shape.restitution = 0.2

        # play a nice sound
        self.snd_ex.set_volume(.4)
        self.snd_ex.play()

        # finally, we add all the individual sprites to the world
        # and remove the bundle, since really, its SUPER DEAD by now.
        [ self.add(s) for s in sb.sprites ]
        self.remove(sb)

        # and make sure its dead (double tap)
        if isinstance(sb, Actor):
            sb.destroy()

        # HACK
        # clear references from the spritebundle
        sb.joints = None
        sb._sprites = None


    # protect some commonly accessed attributes by returning a copy
    @property
    def sprites(self):
        return self._sprites[:]

    def remove(self, obj):
        World.remove(self, obj)
        if obj in self._sprites:
            self._sprites.remove(obj)
            for i in [ k for k,v in self._shapeDB.items() if v == obj ]:
                del self._shapeDB[i]

    def handle_event(self, event):
        World.handle_event(self, event)

        if event.type == MOUSEMOTION:
            pass

        if event.type == KEYDOWN:
            if event.key == K_e:

                # blow up all the spritebundles.  =)
                for sb in [ sb for sb in self._sprites if isinstance(sb, SpriteBundle) ]:
                    self.explode(sb)

    def update(self, time):
        if self._is_enabled:
            # update our "children", or whatever
            World.update(self, time)

            # gets filled with collisions
            self.points = []

            # hackish way of keeping sounds from being duplicated
            played_sounds = []

            # simulate the world some
            self.world.Step(self.timeStep, self.velocityIterations, self.positionIterations)

            # give our sprites (includes actors and bundles) a chance to do something
            [ s.update() for s in self._sprites ]


            # this is not in any way a good method of handling things...
            for p in self.points:  
                if p.state == fwContactTypes.contactPersisted:
                    continue

                body = None
                if p.shape1.isSensor:
                    body = p.shape1.GetBody()

                if p.shape2.isSensor:
                    body = p.shape2.GetBody()

                if body == None:
                    continue

                center = body.GetWorldCenter()

                d = p.position - center

                if d.LengthSquared() < b2d.B2_FLT_EPSILON * b2d.B2_FLT_EPSILON:
                    continue

                d.Normalize()
                F = 5000.0 * d
                body.ApplyForce(F, p.position)


            # handle collisions:
            #   damage and sound is handled here

            for (b0, b1, v) in [(p.shape1, p.shape2, p.velocity) for p in self.points]:
                #if point.state == fwContactTypes.contactPersisted:
                #    continue

                sound = (b0.filter.groupIndex, b1.filter.groupIndex)
                combo = (b0, b1)

                if b0.isSensor: continue
                if b1.isSensor: continue

                if b0.userData == "mouseJoint": continue
                if b1.userData == "mouseJoint": continue

                # attempt to inflict damge on stuff
                try:
                    s = self._shapeDB[b0]
                except KeyError:
                    pass
                else:
                    if isinstance(s, Actor):
                        s.hp -= sqrt(pow(v[0],2)*pow(v[1],2)) / s.toughness
                        if s.hp == 0: self.explode(s)

                try:
                    s = self._shapeDB[b1]
                except KeyError:
                    pass
                else:
                    if isinstance(s, Actor):
                        s.hp -= sqrt(pow(v[0],2)*pow(v[1],2)) / s.toughness
                        if s.hp == 0: self.explode(s)

                if sound in self.stale_sounds : continue

                try:
                    snd = c_sound[sound]
                except KeyError:
                    pass
                else:
                    # super simple way to set volume based on impact velocity
                    vol = sqrt(pow(v[0],2)*pow(v[1],2)) / 5
                    if vol > 1: vol = 1

                    # wish i knew the math to shape this curve to something else
                    # ideally, there would be much less large sounds, but the 
                    # small sounds would be loud as well
                    if vol == 1:
                        snd = self.snd_bump_high[randint(0, len(self.snd_bump_low)-1)]
                    if vol >= 0.99:
                        #snd = self.snd_bump_high[randint(0, len(self.snd_bump_low)-1)]
                        snd = self.snd_bump_low[randint(0, len(self.snd_bump_low)-1)]
                    elif 0.6 <= vol < 0.99:
                        snd = self.snd_bump_low[randint(0, len(self.snd_bump_low)-1)]
                    elif 0.1 <= vol < 0.6:
                        snd = self.snd_bump_low[randint(0, len(self.snd_bump_low)-1)]
                        vol = .1
                    else:
                        continue
                 
                    vol = vol * 0.4
 
                    played_sounds.append(sound) 
                    snd.set_volume(vol)
                    snd.play()

            # remove un-stale collisions
            [ self.stale_sounds.remove(x) for x in self.stale_sounds if x not in played_sounds ]

            # make a list of already handled sounds
            [ self.stale_sounds.append(x) for x in played_sounds if x not in self.stale_sounds ]

    def coordsToMeters(self, *v):
        """
        convert a list pixels to meters, what the engine expects
        accepts a list of coord pairs
        """

        return [ self.toWorld(cp) for cp in v ]

    def toWorld(self, v):
        """
        convert pixels to meters, what the engine expects
        accepts one pair
        """

        return (v[0] / self.ppm, v[1] / self.ppm)

    def translate(self, v):
        return b2d.b2Vec2(self.toWorld((v[0], gfx.screen_dim[1] - v[1])))

    def _add_other(self, obj):
        if isinstance(obj, SpriteBundle):
            self._sprites.append(obj)
            for s in obj.sprites:
                for shape in s.body.GetShapeList():
                    self._shapeDB[shape] = obj

        elif isinstance(obj, PhysicsSprite):
            self._sprites.append(obj)
            for shape in obj.body.GetShapeList():
                self._shapeDB[shape] = obj
        else:
            raise ValueError, "A PhysicsWorld cannot accept type: %s" % type(obj)


class PhysicsWorldCamera(DrawableObject, EventHandler):
    """
    Class to view PhysicsSprites
    They need to be handled slightly differently from normal sprites
    """

    def __init__(self, world):
        DrawableObject.__init__(self)

        self.world = world
        self.ppm = world.ppm
        self.zoom = 1
        self.xoffset = 0
        self.yoffset = 0
        self.extent = None
        self.old_state = (None, None, None)

        self.x_center = None
        self.y_center = None

        self.d = 0
        self.tz = self.zoom
        self.zt = 0.0
        self.zs = .5

        # pan with an object with it in the center of the cam
        self.obj_tracking = None

    def handle_event(self, event):
        if event.type == KEYDOWN:
            if event.key == K_o:
                self.tz -= .1
                self.zt = 0.0
            elif event.key == K_p:
                self.tz += .1
                self.zt = 0.0
        return event

    def update(self, time):
        if self.tz < self.zoom:
            if self.d != 1:
                self.zt = 0.0
                self.d = 1
            else:
                if self.zt < 5: self.zt += 0.1
            self.zoom -= pow(self.zs, self.zt) / 100.0

        elif self.tz > self.zoom:
            if self.d != 2:
                self.zt = 0.0
                self.d = 1
            else:
                if self.zt < 5: self.zt += 0.1
            self.zoom += pow(self.zs, self.zt) / 100.0

        if round(self.zoom, 2) == self.tz:
            self.d = 0
            self.zt = 0.0
            self.zoom = self.tz
            self.tz = self.zoom

        #x = self.zt
        #cosxpi = cos(x * pi)
        #y = .5 * (a * (cosxpi - 1) + b * (1 - cosxpi))
        #print y

        #print self.tz, self.zoom, self.zt

    def draw(self, surface):
        """
        Draw a group of sprites onto a surface
        """

        self.x_size, self.y_size = surface.get_rect().size

        if (self.zoom, self.xoffset, self.yoffset) != self.old_state:
            if self.old_state[0] != self.zoom:
                if (self.x_center == None) or (self.y_center == None):
                    self.x_center = self.x_size / 2
                    self.y_center = self.y_size / 2
                if self.zoom != 0:
                    self.xoffset = self.x_center + (self.x_size/2) - (self.x_size / self.zoom) 
                    self.yoffset = self.y_center + (self.y_size/2) - (self.y_size / self.zoom)
                else:
                    self.xoffset = self.x_center - self.x_size / 2
                    self.yoffset = self.y_center - self.y_size / 2

            if self.zoom != 0:
                self.extent = Rect((1,1,1,1)).inflate(self.x_size / self.zoom, self.y_size / self.zoom)
            else:
                self.extent = Rect((1,1,1,1)).inflate(self.x_size, self.y_size)
            self.extent.topleft = (self.xoffset, self.yoffset)
            self.old_state = (self.zoom, self.xoffset, self.yoffset)

        sprites = self.world.sprites

        # TODO: this will update (and transfor rects even if they are off-screen
        [ s.update_rect(self) for s in sprites ]

        # BUG: Doesn't work...well.  sprites get lost near edges of screen
        #[ surface.blit(s.image, s.rect) for s in sprites if self.extent.colliderect(s.rect) ]
        #[ surface.blit(s.image, s.rect) for s in sprites if self.extent.contains(s.rect) ]
        [ surface.blit(s.image, s.rect) for s in sprites ]

        if dbg_outlines:
            for s in sprites:
                # BUG: SpriteBundles will not have their lines drawn
                try:
                    body = s.body
                except AttributeError:
                    continue
                xform = body.GetXForm()
                shape = body.GetShapeList()[0]

                points = [ self.toSurface(b2d.b2Mul(xform, v)) for v in shape.vertices ]
                aalines(surface, (0,255,100), True, points, 1)

    def toSurface(self, v):
        """
        convert a box2d vertex to surfacespace, accounting for zoom
        """

        return ((v[0]*self.ppm*self.zoom) - self.xoffset, 
                self.y_size - ((v[1]*self.ppm*self.zoom) - self.yoffset))

    def toWorld(self, v):
        try:
            return b2d.b2Vec2((v[0] + self.xoffset) / self.ppm / self.zoom,
                    (self.y_size - v[1] + self.yoffset) / self.ppm / self.zoom)
        except:
            return None
