"""
Copyright 2011,  Leif Theden & Michael Bachmann


This file is part of Nein.

Nein is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Nein is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Nein.  If not, see <http://www.gnu.org/licenses/>.
"""

import pygame
import Box2D as b2d
from pygame.rect import Rect
from pygame.surface import Surface
from pygame.transform import rotozoom
from pygame.sprite import Sprite
from math import degrees

from engine import DrawableObject, UpdateableObject


# if this is greater than 0, then rotations will be
# limited to that number of angles in a 360 cicrle
simple_no = 128
simple_angle = 360.0 / simple_no

slow_pc = True

class PhysicsSprite(Sprite):
    """
    Hack to use pygame and pyBox2d together
    """

    pad_alpha = True

    def __init__(self, image, body):
        super(PhysicsSprite, self).__init__()

        if PhysicsSprite.pad_alpha:
            image = self.padimage(image)

        self.body = body
        self.normal_image = image

        # these will be created when the sprite is updated
        self.image = None
        self.rect = None

        # prevents uneeded transformations
        self.old_trans = (None, None)

    def handle_collision(self, other, velocity):
        pass

    def padimage(self, image):    
        """
        Do a little processing of the input image to make it purdyer.
        Pad the image with transparent pixels so the edged get antialised
        when rotated and scaled.
        """

        # replacement surface that is slightly bigger then the original
        new = Surface(image.get_rect().inflate(2, 2).size, pygame.SRCALPHA)

        # get the color on the border
        color = image.get_at((0,0))

        # set the color we want to be slightly opaque
        color[3] = 0

        # fill the replacement surface
        new.fill(color)

        new.blit(image, (1,1))

        return new

    def update_rect(self, camera):
        """
        position the rect to whatever the physics engine says it should be
        updates the image and rect based on rotation of the physics sprite
        """

        # TODO: for low end computers, this number can be rounded to
        # reduce the number of rotations required to display the object.

        if slow_pc:
            angle = round(int(degrees(self.body.GetAngle()) % 360.0 / 360.0 * simple_no) * simple_angle, 2)
        else:
            angle = degrees(self.body.GetAngle()) % 360.0

        if (angle, camera.zoom) != self.old_trans:
            old_trans = (angle, camera.zoom)
            self.image = rotozoom(self.normal_image, angle, camera.zoom)
            self.rect = self.image.get_rect()

        # kinda a hackish way to get the rect here
        xform = self.body.GetXForm()
        shape = self.body.GetShapeList()[0]

        # boxes
        if shape.GetType() == 1:
            points = [ camera.toSurface(b2d.b2Mul(xform, v)) for v in shape.vertices ]

            # hack to get the upper left corner
            min_x = min([ x[0] for x in points ])
            min_y = min([ x[1] for x in points ])
            self.rect.topleft = (min_x, min_y)

        # circles
        elif shape.GetType() == 0:
            self.rect.center = camera.toSurface(self.body.GetWorldCenter())

class SpriteBundle(PhysicsSprite):
    """
    Used for ragdolls.
    This class is a collection of PhysicsSprites that is logically one object in the game.
    Stores the sprites and references to the joints that hold it together.
    Rendered as one object.
    """

    def __init__(self, sprites, joints):
        self._sprites = sprites
        self.joints = joints

        # these will be [hopefully] generated before the first draw
        self.rect = Rect(0,0,0,0)
        self.image = None

    # protect some commonly accessed attributes by returning a copy
    @property
    def sprites(self):
        return self._sprites[:]

    def update_rect(self, camera):
        """
        kinda slow, but works for now, we create a surface, blit our individual
        sprites to it and store it later for our parent to blit.
        """

        # update our sprites, and make a bounding box that fits them all
        [ s.update_rect(camera) for s in self._sprites ] 
        self.rect = self._sprites[0].rect.unionall([ s.rect for s in self._sprites[1:] ])

        # make a new surface for ourselves, then blit the sprites
        # the sprites rect is moved so they fit in the new surface
        image = Surface(self.rect.size, pygame.SRCALPHA)
        xoffset, yoffset = (-self.rect.left, -self.rect.top)
        [ image.blit(s.image, s.rect.move(xoffset, yoffset)) for s in self._sprites ]

        # TODO: it would make sence to make the surface a little bigger than needed
        #       and hope that the blits fit into it, so it doesn't have to be created
        #       so often.

        self.image = image 
