"""
Copyright 2011,  Leif Theden & Michael Bachmann


This file is part of Nein.

Nein is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Nein is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Nein.  If not, see <http://www.gnu.org/licenses/>.
"""

from pygame.locals import *
from pygame.transform import scale
from lib.engine import World, GameState
from lib.physicsworld import PhysicsWorldCamera
from lib.inputs import GfxCursor
from pygame.time import get_ticks
import pygame.mouse as mouse
import lib.gui as gui
import lib.sfx as sfx
import lib.gfx as gfx
import Box2D as box2d
from resources.highscore import HighScoreScreen

# get out level: 
from lib.level import *
from lib.buildutils import *

cat0 = None

def build_world():
    global cat0
    from lib.physicsworld import PhysicsWorld

    world = PhysicsWorld()

    world.add(trampoline(world, (10,10)))
    world.add(slugger(world, (400,120)))
    world.add(beachball(world, (300, 200)))

    cat0 = make_cat_ragdoll(world, (350,230))
    world.add(cat0)

    image = gfx.load("tinycat.png")
    cat1 = make_sprite(world, image, density=999, pos=(200, 90))
    cat1.body.mass = 999
    world.add(cat1)

    image = pygame.surface.Surface((800, 2))
    image.fill((0,0,0))
    ground = static_sprite(world, image, pos=(0,0))
    world.add(ground)

    ceiling = static_sprite(world, image, pos=(0,300))
    world.add(ceiling)

    image = pygame.surface.Surface((2, 300))
    image.fill((100,10,20))
    left = static_sprite(world, image, pos=(0,0))
    world.add(left)

    right = static_sprite(world, image, pos=(500, 0))
    world.add(right)

    return world

class Level1(GameState):
    def __init__(self, driver):
        GameState.__init__(self, driver)

        world = build_world()
        self.camera0 = PhysicsWorldCamera(world)
        world.add(self.camera0)
        sprite_world = World()
        sprite_world.add(world)

        self.cr_open = scale(gfx.load_colorkey("open.png").convert(), (20,30))
        self.cr_grasp = scale(gfx.load_colorkey("grasp.png").convert(), (20, 25))

        self.cursor = GfxCursor(None)
        self.cursor.setCursor(self.cr_open, (7,8))

        # build the "level overlay"
        hp_bar = gui.HorizBar((0,0,500,10), func=cat0.get_hp)
        sprite_world.add(hp_bar)

        self.sprite_world = sprite_world
        self.pworld = world
        self.world = world.world

        self.mouseJoint = None

        self.fadeout = False

    def build_world(self):
        pass

    def activate(self):
        gfx.set_pixelize(True)
        self._driver.reloadScreen()
        sfx.load_music("EPGame1-lofi.ogg")
        sfx.play_music(-1)
        self.cursor.disable() 
        r = gfx.screen.get_rect()
        mouse.set_pos(r.width / 2, r.height / 2)
        self.cursor.enable()
        self.start = None

    def deactivate(self):
        self.cursor.disable()
        gfx.set_pixelize(False)
        self._driver.reloadScreen()

    def handle_event(self, event):
        if event.type == MOUSEMOTION:
            pos = (event.pos[0] / gfx.pix_scale, event.pos[1] / gfx.pix_scale)
            self.cursor.update(pos)
            self.mousemotion(self.camera0.toWorld(pos))
        if event.type == MOUSEBUTTONUP:
            pos = (event.pos[0] / gfx.pix_scale, event.pos[1] / gfx.pix_scale)
            self.cursor.setCursor(self.cr_open, (7,8))
            self.mouseup(self.camera0.toWorld(pos))
        if event.type == MOUSEBUTTONDOWN:
            pos = (event.pos[0] / gfx.pix_scale, event.pos[1] / gfx.pix_scale)
            self.cursor.setCursor(self.cr_grasp, (7,8))
            self.mousedown(self.camera0.toWorld(pos))

        self.sprite_world.handle_event(event)

    def draw(self, surface):
        # a transition into the next screen
        if self.fadeout:
            x = int((get_ticks()-self.start) / 1500.0 * gfx.buffer_dim[0])
            surface.fill((0,0,0), (0,0,x,gfx.buffer_dim[1]))
            self.t_oldx = x
            if x >= gfx.buffer_dim[0]:
                self._driver.push(Level1(self._driver))
                self._driver.replace(HighScoreScreen(self._driver))
            return

        surface.fill((100,100,100))
        self.sprite_world.draw(surface)
        self.cursor.draw(surface)

    def update(self, time):
        self.sprite_world.update(time)

        if cat0.is_living == False:
            if self.start == None:
                self.start = get_ticks()
            elif self.fadeout == False:
                if get_ticks() - self.start > 3000:
                    self.fadeout = True
                    pygame.mixer.music.fadeout(1500)
                    

    def mousedown(self, p):
        if self.mouseJoint != None:
            return

        # Create a mouse joint on the selected body (assuming it's dynamic)

        # Make a small box.
        aabb = box2d.b2AABB()
        aabb.lowerBound = p - (0.1, 0.1)
        aabb.upperBound = p + (0.1, 0.1)

        # Query the world for overlapping shapes.
        body = None
        k_maxCount = 10 # maximum amount of shapes to return

        (count, shapes) = self.world.Query(aabb, k_maxCount)
        for shape in shapes:
            shapeBody = shape.GetBody()
            if not shapeBody.IsStatic() and shapeBody.GetMass() > 0.0:
                if shape.TestPoint(shapeBody.GetXForm(), p): # is it inside?
                    body = shapeBody
                    break
        
        if body:
            # A body was selected, create the mouse joint
            md = box2d.b2MouseJointDef()
            md.body1   = self.world.GetGroundBody()
            md.body2   = body
            md.target  = p
            md.maxForce= 2000.0 * body.GetMass()
            self.mouseJoint = self.world.CreateJoint(md)
            body.WakeUp()

    def mouseup(self, p):
        if self.mouseJoint:
            self.world.DestroyJoint(self.mouseJoint)
            self.mouseJoint = None

    def mousemotion(self, p):
        self.mouseWorld = p
        if self.mouseJoint:
            self.mouseJoint.SetTarget(p)
